package org.eclipse.jdt.internal.compiler.apt.model;

package .tmp;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.AnnotationValue;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Name;
import javax.lang.model.element.PackageElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.util.Elements;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.apt.dispatch.BaseProcessingEnvImpl;
import org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl;
import org.eclipse.jdt.internal.compiler.apt.model.ElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.ExecutableElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.Factory;
import org.eclipse.jdt.internal.compiler.apt.model.NameImpl;
import org.eclipse.jdt.internal.compiler.apt.model.PackageElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.TypeElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.VariableElementImpl;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Javadoc;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;

public class ElementsImpl implements Elements {
  private static final Pattern INITIAL_DELIMITER = Pattern.compile("^\\s*/\\*+");
  
  private final BaseProcessingEnvImpl _env;
  
  public ElementsImpl(BaseProcessingEnvImpl env) {
    this._env = env;
  }
  
  public List<? extends AnnotationMirror> getAllAnnotationMirrors(Element e) {
    if (e.getKind() == ElementKind.CLASS && e instanceof TypeElementImpl) {
      List<AnnotationBinding> annotations = new ArrayList<>();
      Set<ReferenceBinding> annotationTypes = new HashSet<>();
      ReferenceBinding binding = (ReferenceBinding)((TypeElementImpl)e)._binding;
      boolean checkIfInherited = false;
      while (binding != null) {
        if (binding instanceof ParameterizedTypeBinding)
          binding = ((ParameterizedTypeBinding)binding).genericType(); 
        byte b;
        int i;
        AnnotationBinding[] arrayOfAnnotationBinding;
        for (i = (arrayOfAnnotationBinding = Factory.getPackedAnnotationBindings(binding.getAnnotations())).length, b = 0; b < i; ) {
          AnnotationBinding annotation = arrayOfAnnotationBinding[b];
          if (annotation != null) {
            ReferenceBinding annotationType = annotation.getAnnotationType();
            if (!checkIfInherited || (annotationType.getAnnotationTagBits() & 0x1000000000000L) != 0L)
              if (!annotationTypes.contains(annotationType)) {
                annotationTypes.add(annotationType);
                annotations.add(annotation);
              }  
          } 
          b++;
        } 
        binding = binding.superclass();
        checkIfInherited = true;
      } 
      List<AnnotationMirror> list = new ArrayList<>(annotations.size());
      for (AnnotationBinding annotation : annotations)
        list.add(this._env.getFactory().newAnnotationMirror(annotation)); 
      return Collections.unmodifiableList(list);
    } 
    return e.getAnnotationMirrors();
  }
  
  public List<? extends Element> getAllMembers(TypeElement type) {
    if (type == null || !(type instanceof TypeElementImpl))
      return Collections.emptyList(); 
    ReferenceBinding binding = (ReferenceBinding)((TypeElementImpl)type)._binding;
    Map<String, ReferenceBinding> types = new HashMap<>();
    List<FieldBinding> fields = new ArrayList<>();
    Map<String, Set<MethodBinding>> methods = new HashMap<>();
    Set<ReferenceBinding> superinterfaces = new LinkedHashSet<>();
    boolean ignoreVisibility = true;
    while (binding != null) {
      addMembers(binding, ignoreVisibility, types, fields, methods);
      Set<ReferenceBinding> newfound = new LinkedHashSet<>();
      collectSuperInterfaces(binding, superinterfaces, newfound);
      for (ReferenceBinding superinterface : newfound)
        addMembers(superinterface, false, types, fields, methods); 
      superinterfaces.addAll(newfound);
      binding = binding.superclass();
      ignoreVisibility = false;
    } 
    List<Element> allMembers = new ArrayList<>();
    for (ReferenceBinding nestedType : types.values())
      allMembers.add(this._env.getFactory().newElement((Binding)nestedType)); 
    for (FieldBinding field : fields)
      allMembers.add(this._env.getFactory().newElement((Binding)field)); 
    for (Set<MethodBinding> sameNamedMethods : methods.values()) {
      for (MethodBinding method : sameNamedMethods)
        allMembers.add(this._env.getFactory().newElement((Binding)method)); 
    } 
    return allMembers;
  }
  
  private void collectSuperInterfaces(ReferenceBinding type, Set<ReferenceBinding> existing, Set<ReferenceBinding> newfound) {
    byte b;
    int i;
    ReferenceBinding[] arrayOfReferenceBinding;
    for (i = (arrayOfReferenceBinding = type.superInterfaces()).length, b = 0; b < i; ) {
      ReferenceBinding superinterface = arrayOfReferenceBinding[b];
      if (!existing.contains(superinterface) && !newfound.contains(superinterface)) {
        newfound.add(superinterface);
        collectSuperInterfaces(superinterface, existing, newfound);
      } 
      b++;
    } 
  }
  
  private void addMembers(ReferenceBinding binding, boolean ignoreVisibility, Map<String, ReferenceBinding> types, List<FieldBinding> fields, Map<String, Set<MethodBinding>> methods) {
    byte b;
    int i;
    ReferenceBinding[] arrayOfReferenceBinding;
    for (i = (arrayOfReferenceBinding = binding.memberTypes()).length, b = 0; b < i; ) {
      ReferenceBinding subtype = arrayOfReferenceBinding[b];
      if (ignoreVisibility || !subtype.isPrivate()) {
        String name = new String(subtype.sourceName());
        if (types.get(name) == null)
          types.put(name, subtype); 
      } 
      b++;
    } 
    FieldBinding[] arrayOfFieldBinding;
    for (i = (arrayOfFieldBinding = binding.fields()).length, b = 0; b < i; ) {
      FieldBinding field = arrayOfFieldBinding[b];
      if (ignoreVisibility || !field.isPrivate())
        fields.add(field); 
      b++;
    } 
    MethodBinding[] arrayOfMethodBinding;
    for (i = (arrayOfMethodBinding = binding.methods()).length, b = 0; b < i; ) {
      MethodBinding method = arrayOfMethodBinding[b];
      if (!method.isSynthetic() && (ignoreVisibility || (!method.isPrivate() && !method.isConstructor()))) {
        String methodName = new String(method.selector);
        Set<MethodBinding> sameNamedMethods = methods.get(methodName);
        if (sameNamedMethods == null) {
          sameNamedMethods = new HashSet<>(4);
          methods.put(methodName, sameNamedMethods);
          sameNamedMethods.add(method);
        } else {
          boolean unique = true;
          if (!ignoreVisibility)
            for (MethodBinding existing : sameNamedMethods) {
              MethodVerifier verifier = this._env.getLookupEnvironment().methodVerifier();
              if (verifier.doesMethodOverride(existing, method)) {
                unique = false;
                break;
              } 
            }  
          if (unique)
            sameNamedMethods.add(method); 
        } 
      } 
      b++;
    } 
  }
  
  public Name getBinaryName(TypeElement type) {
    TypeElementImpl typeElementImpl = (TypeElementImpl)type;
    ReferenceBinding referenceBinding = (ReferenceBinding)typeElementImpl._binding;
    return (Name)new NameImpl(
        CharOperation.replaceOnCopy(referenceBinding.constantPoolName(), '/', '.'));
  }
  
  public String getConstantExpression(Object value) {
    if (!(value instanceof Integer) && 
      !(value instanceof Byte) && 
      !(value instanceof Float) && 
      !(value instanceof Double) && 
      !(value instanceof Long) && 
      !(value instanceof Short) && 
      !(value instanceof Character) && 
      !(value instanceof String) && 
      !(value instanceof Boolean))
      throw new IllegalArgumentException("Not a valid wrapper type : " + value.getClass()); 
    if (value instanceof Character) {
      StringBuilder builder = new StringBuilder();
      builder.append('\'').append(value).append('\'');
      return String.valueOf(builder);
    } 
    if (value instanceof String) {
      StringBuilder builder = new StringBuilder();
      builder.append('"').append(value).append('"');
      return String.valueOf(builder);
    } 
    if (value instanceof Float) {
      StringBuilder builder = new StringBuilder();
      builder.append(value).append('f');
      return String.valueOf(builder);
    } 
    if (value instanceof Long) {
      StringBuilder builder = new StringBuilder();
      builder.append(value).append('L');
      return String.valueOf(builder);
    } 
    if (value instanceof Short) {
      StringBuilder builder = new StringBuilder();
      builder.append("(short)").append(value);
      return String.valueOf(builder);
    } 
    if (value instanceof Byte) {
      StringBuilder builder = new StringBuilder();
      builder.append("(byte)0x");
      int intValue = ((Byte)value).byteValue();
      String hexString = Integer.toHexString(intValue & 0xFF);
      if (hexString.length() < 2)
        builder.append('0'); 
      builder.append(hexString);
      return String.valueOf(builder);
    } 
    return String.valueOf(value);
  }
  
  public String getDocComment(Element e) {
    char[] unparsed = getUnparsedDocComment(e);
    return formatJavadoc(unparsed);
  }
  
  private char[] getUnparsedDocComment(Element e) {
    TypeDeclaration typeDeclaration;
    TypeElementImpl typeElementImpl;
    ReferenceBinding referenceBinding;
    PackageElementImpl packageElementImpl;
    PackageBinding packageBinding;
    char[][] compoundName;
    ReferenceBinding type;
    ExecutableElementImpl executableElementImpl;
    MethodBinding methodBinding;
    AbstractMethodDeclaration sourceMethod;
    VariableElementImpl variableElementImpl;
    FieldBinding fieldBinding;
    FieldDeclaration sourceField;
    Javadoc javadoc = null;
    ReferenceContext referenceContext = null;
    switch (e.getKind()) {
      case ENUM:
      case CLASS:
      case null:
      case INTERFACE:
        typeElementImpl = (TypeElementImpl)e;
        referenceBinding = (ReferenceBinding)typeElementImpl._binding;
        if (referenceBinding instanceof SourceTypeBinding) {
          SourceTypeBinding sourceTypeBinding = (SourceTypeBinding)referenceBinding;
          typeDeclaration = sourceTypeBinding.scope.referenceContext;
          javadoc = typeDeclaration.javadoc;
        } 
        break;
      case PACKAGE:
        packageElementImpl = (PackageElementImpl)e;
        packageBinding = (PackageBinding)packageElementImpl._binding;
        compoundName = CharOperation.arrayConcat(packageBinding.compoundName, TypeConstants.PACKAGE_INFO_NAME);
        type = this._env.getLookupEnvironment().getType(compoundName);
        if (type != null && type.isValidBinding() && type instanceof SourceTypeBinding) {
          SourceTypeBinding sourceTypeBinding = (SourceTypeBinding)type;
          typeDeclaration = sourceTypeBinding.scope.referenceContext;
          javadoc = typeDeclaration.javadoc;
        } 
        break;
      case METHOD:
      case CONSTRUCTOR:
        executableElementImpl = (ExecutableElementImpl)e;
        methodBinding = (MethodBinding)executableElementImpl._binding;
        sourceMethod = methodBinding.sourceMethod();
        if (sourceMethod != null) {
          javadoc = sourceMethod.javadoc;
          AbstractMethodDeclaration abstractMethodDeclaration = sourceMethod;
        } 
        break;
      case ENUM_CONSTANT:
      case FIELD:
        variableElementImpl = (VariableElementImpl)e;
        fieldBinding = (FieldBinding)variableElementImpl._binding;
        sourceField = fieldBinding.sourceField();
        if (sourceField != null) {
          javadoc = sourceField.javadoc;
          if (fieldBinding.declaringClass instanceof SourceTypeBinding) {
            SourceTypeBinding sourceTypeBinding = (SourceTypeBinding)fieldBinding.declaringClass;
            typeDeclaration = sourceTypeBinding.scope.referenceContext;
          } 
        } 
        break;
      default:
        return null;
    } 
    if (javadoc != null && typeDeclaration != null) {
      char[] contents = typeDeclaration.compilationResult().getCompilationUnit().getContents();
      if (contents != null)
        return CharOperation.subarray(contents, javadoc.sourceStart, javadoc.sourceEnd - 1); 
    } 
    return null;
  }
  
  private static String formatJavadoc(char[] unparsed) {
    if (unparsed == null || unparsed.length < 5)
      return null; 
    String[] lines = (new String(unparsed)).split("\n");
    Matcher delimiterMatcher = INITIAL_DELIMITER.matcher(lines[0]);
    if (!delimiterMatcher.find())
      return null; 
    int iOpener = delimiterMatcher.end();
    lines[0] = lines[0].substring(iOpener);
    if (lines.length == 1) {
      StringBuilder stringBuilder = new StringBuilder();
      char[] chars = lines[0].toCharArray();
      boolean startingWhitespaces = true;
      byte b;
      int i;
      char[] arrayOfChar1;
      for (i = (arrayOfChar1 = chars).length, b = 0; b < i; ) {
        char c = arrayOfChar1[b];
        if (Character.isWhitespace(c)) {
          if (!startingWhitespaces)
            stringBuilder.append(c); 
        } else {
          startingWhitespaces = false;
          stringBuilder.append(c);
        } 
        b++;
      } 
      return stringBuilder.toString();
    } 
    int firstLine = (lines[0].trim().length() > 0) ? 0 : 1;
    int lastLine = (lines[lines.length - 1].trim().length() > 0) ? (lines.length - 1) : (lines.length - 2);
    StringBuilder sb = new StringBuilder();
    if (lines[0].length() != 0 && firstLine == 1)
      sb.append('\n'); 
    boolean preserveLineSeparator = (lines[0].length() == 0);
    for (int line = firstLine; line <= lastLine; line++) {
      char[] chars = lines[line].toCharArray();
      int starsIndex = getStars(chars);
      int leadingWhitespaces = 0;
      boolean recordLeadingWhitespaces = true;
      for (int i = 0, max = chars.length; i < max; i++) {
        char c = chars[i];
        switch (c) {
          case ' ':
            if (starsIndex == -1) {
              if (recordLeadingWhitespaces) {
                leadingWhitespaces++;
                break;
              } 
              sb.append(c);
              break;
            } 
            if (i >= starsIndex)
              sb.append(c); 
            break;
          default:
            recordLeadingWhitespaces = false;
            if (leadingWhitespaces != 0) {
              int numberOfTabs = leadingWhitespaces / 8;
              if (numberOfTabs != 0) {
                for (int j = 0, max2 = numberOfTabs; j < max2; j++)
                  sb.append("        "); 
                if (leadingWhitespaces % 8 >= 1)
                  sb.append(' '); 
              } else if (line != 0) {
                for (int j = 0, max2 = leadingWhitespaces; j < max2; j++)
                  sb.append(' '); 
              } 
              leadingWhitespaces = 0;
              sb.append(c);
              break;
            } 
            if (c == '\t') {
              if (i >= starsIndex)
                sb.append(c); 
              break;
            } 
            if (c != '*' || i > starsIndex)
              sb.append(c); 
            break;
        } 
      } 
      int end = lines.length - 1;
      if (line < end) {
        sb.append('\n');
      } else if (preserveLineSeparator && line == end) {
        sb.append('\n');
      } 
    } 
    return sb.toString();
  }
  
  private static int getStars(char[] line) {
    for (int i = 0, max = line.length; i < max; i++) {
      char c = line[i];
      if (!Character.isWhitespace(c)) {
        if (c == '*') {
          for (int j = i + 1; j < max; j++) {
            if (line[j] != '*')
              return j; 
          } 
          return max - 1;
        } 
        break;
      } 
    } 
    return -1;
  }
  
  public Map<? extends ExecutableElement, ? extends AnnotationValue> getElementValuesWithDefaults(AnnotationMirror a) {
    return ((AnnotationMirrorImpl)a).getElementValuesWithDefaults();
  }
  
  public Name getName(CharSequence cs) {
    return (Name)new NameImpl(cs);
  }
  
  public PackageElement getPackageElement(CharSequence name) {
    LookupEnvironment le = this._env.getLookupEnvironment();
    if (name.length() == 0)
      return (PackageElement)new PackageElementImpl(this._env, le.defaultPackage); 
    char[] packageName = name.toString().toCharArray();
    PackageBinding packageBinding = le.createPackage(CharOperation.splitOn('.', packageName));
    if (packageBinding == null)
      return null; 
    return (PackageElement)new PackageElementImpl(this._env, packageBinding);
  }
  
  public PackageElement getPackageOf(Element type) {
    TypeElementImpl typeElementImpl;
    ReferenceBinding referenceBinding;
    ExecutableElementImpl executableElementImpl;
    MethodBinding methodBinding;
    VariableElementImpl variableElementImpl;
    FieldBinding fieldBinding;
    LocalVariableBinding localVariableBinding;
    switch (type.getKind()) {
      case ENUM:
      case CLASS:
      case null:
      case INTERFACE:
        typeElementImpl = (TypeElementImpl)type;
        referenceBinding = (ReferenceBinding)typeElementImpl._binding;
        return (PackageElement)this._env.getFactory().newElement((Binding)referenceBinding.fPackage);
      case PACKAGE:
        return (PackageElement)type;
      case METHOD:
      case CONSTRUCTOR:
        executableElementImpl = (ExecutableElementImpl)type;
        methodBinding = (MethodBinding)executableElementImpl._binding;
        return (PackageElement)this._env.getFactory().newElement((Binding)methodBinding.declaringClass.fPackage);
      case ENUM_CONSTANT:
      case FIELD:
        variableElementImpl = (VariableElementImpl)type;
        fieldBinding = (FieldBinding)variableElementImpl._binding;
        return (PackageElement)this._env.getFactory().newElement((Binding)fieldBinding.declaringClass.fPackage);
      case PARAMETER:
        variableElementImpl = (VariableElementImpl)type;
        localVariableBinding = (LocalVariableBinding)variableElementImpl._binding;
        return (PackageElement)this._env.getFactory().newElement((Binding)(localVariableBinding.declaringScope.classScope()).referenceContext.binding.fPackage);
      case LOCAL_VARIABLE:
      case EXCEPTION_PARAMETER:
      case STATIC_INIT:
      case INSTANCE_INIT:
      case TYPE_PARAMETER:
      case OTHER:
        return null;
    } 
    return null;
  }
  
  public TypeElement getTypeElement(CharSequence name) {
    LookupEnvironment le = this._env.getLookupEnvironment();
    char[][] compoundName = CharOperation.splitOn('.', name.toString().toCharArray());
    ReferenceBinding binding = le.getType(compoundName);
    if (binding == null) {
      ReferenceBinding topLevelBinding = null;
      int topLevelSegments = compoundName.length;
      while (--topLevelSegments > 0) {
        char[][] topLevelName = new char[topLevelSegments][];
        for (int j = 0; j < topLevelSegments; j++)
          topLevelName[j] = compoundName[j]; 
        topLevelBinding = le.getType(topLevelName);
        if (topLevelBinding != null)
          break; 
      } 
      if (topLevelBinding == null)
        return null; 
      binding = topLevelBinding;
      for (int i = topLevelSegments; binding != null && i < compoundName.length; i++)
        binding = binding.getMemberType(compoundName[i]); 
    } 
    if (binding == null)
      return null; 
    return (TypeElement)new TypeElementImpl(this._env, binding, null);
  }
  
  public boolean hides(Element hider, Element hidden) {
    if (hidden == null)
      throw new NullPointerException(); 
    return ((ElementImpl)hider).hides(hidden);
  }
  
  public boolean isDeprecated(Element e) {
    if (!(e instanceof ElementImpl))
      return false; 
    return ((((ElementImpl)e)._binding.getAnnotationTagBits() & 0x400000000000L) != 0L);
  }
  
  public boolean overrides(ExecutableElement overrider, ExecutableElement overridden, TypeElement type) {
    if (overridden == null || type == null)
      throw new NullPointerException(); 
    return ((ExecutableElementImpl)overrider).overrides(overridden, type);
  }
  
  public void printElements(Writer w, Element... elements) {
    String lineSeparator = System.getProperty("line.separator");
    byte b;
    int i;
    Element[] arrayOfElement;
    for (i = (arrayOfElement = elements).length, b = 0; b < i; ) {
      Element element = arrayOfElement[b];
      try {
        w.write(element.toString());
        w.write(lineSeparator);
      } catch (IOException iOException) {}
      b++;
    } 
    try {
      w.flush();
    } catch (IOException iOException) {}
  }
  
  public boolean isFunctionalInterface(TypeElement type) {
    if (type != null && type.getKind() == ElementKind.INTERFACE) {
      ReferenceBinding binding = (ReferenceBinding)((TypeElementImpl)type)._binding;
      if (binding instanceof SourceTypeBinding)
        return binding.isFunctionalInterface((Scope)((SourceTypeBinding)binding).scope); 
    } 
    return false;
  }
}
